from ..expressions import *
from ..variables import *

"""
Flattening a model (or individual constraints) into a normal form. See docs/behind_the_scenes.srt

- flatten_model(model): flattens constraints and objective, returns new model
- flatten_constraint(constraint): flattens the constraint, returns a list of base constraints

THIS IS ONLY A POTENTIAL STRUCTURE, not tested or run...
"""

def flatten_model(orig_model):
    from ..model import Model # otherwise circular dependency...

    """
        Receives model, returns new model where every constraint is a 'base' constraint
    """
    # the top-level constraints
    basecons = []
    for con in orig_model.constraints:
        basecons += flatten_constraint(con)

    # the objective
    if orig_model.objective is None:
        return Model(basecons) # no objective, satisfaction problem
    else:
        (newobj, newcons) = flatten_objective(orig_model.objective)
        basecons += newcons
        if orig_model.objective_max:
            return Model(basecons, maximize=newobj)
        else:
            return Model(basecons, minimize=newobj)

    return new_model


def flatten_constraint(con):
    """
        input is any expression; except is_num(), pure NumVarImpl, Operator with is_type_num() and Element with len(args)=2
        output is a list of base constraints, each base constraint is one of:
            * BoolVar
            * Operator with is_type_bool(), all args are: BoolVar
            * Operator '->' with args [boolexpr,BoolVar]
            * Operator '->' with args [Comparison,BoolVar]
            * Comparison, all args are is_num() or NumVar 
            * Comparison '==' with args [boolexpr,BoolVar]
            * Comparison '!=' with args [boolexpr,BoolVar]
            * Comparison '==' with args [numexpr,NumVar]
            * Comparison '!=' with args [numexpr,NumVar]
            * Element with len(args)==3 and args = ([NumVar], NumVar, NumVar)
            * Global, all args are: NumVar

        will return 'Error' if something is not supported
        TODO, what built-in python error is best?
    """
    # base cases
    if isinstance(con, BoolVarImpl) or isinstance(con, bool):
        return [con]
    elif is_num(con) or isinstance(con, NumVarImpl):
        raise Exception("Numeric constants or numeric variables not allowed as base constraint")

    basecons = []

    # recursively flatten list of constraints
    if is_any_list(con):
        for con_x in con:
            basecons += flatten_constraint(con_x)
        return basecons

    if isinstance(con, Operator):
        # only Boolean operators allowed as top-level constraint
        # bool: 'and'/n, 'or'/n, 'xor'/n, '->'/2
        if not con.is_bool():
            raise Exception("Operator '{}' not allowed as base constraint".format(expr.name))

        return [con] # TODO

    elif isinstance(con, Comparison):
        #allowed = {'==', '!=', '<=', '<', '>=', '>'}
        return [con] # TODO
        for lvar, rvar in zipcycle(args[0], args[1]):
            if expr.name == '==' or expr.name == '!=':
                # special case... allows some nesting of LHS
                # and a variable on RHS
                # check whether needs swap on LHS...
                return [con] # TODO
            else: # inequalities '<=', '<', '>=', '>'
                # special case... allows some nesting of LHS
                return [con] # TODO

    elif isinstance(con, Element):
        return [con] # TODO

    # rest: global constraints
    else:
        return [con] # TODO

def flatten_objective(expr):
    """
        input: expression of type:
            * is_num()
            * NumVarImpl
            * Operator with is_type_num()
            * Element with len(args)==2
        output: tuple (base_expr, base_cons) with:
            base_expr one of:
                * is_num()
                * NumVarImpl
                * Operator with is_type_num(), all args are: is_num() or NumVarImpl
                * Operator 'sum', all args are: is_num() or NumVarImpl or Operator '*'[is_num(), NumVarImpl]
                * Element with len(args)==2 and args = ([NumVar], NumVar)
            base_cons: list of flattened constraints (with flatten_constraint(con))
    """
    if is_num(expr) or isinstance(expr, NumVarImpl):
        return (expr, [])

    basecons = []
    if isinstance(expr, Operator):
        # only Numeric operators allowed
        if expr.is_bool():
            raise Exception("Boolean operator '{}' not allowed in objective".format(expr.name)) # or bug

        flat_args = [flatten_numexpr(e) for e in expr.args]
        if all(arg is flatarg[0] for (arg,flatarg) in zip(expr.args, flat_args)):
            return (expr, [])
        else:
            # one of the args was changed
            newargs = [flatarg[0] for flatarg in flat_args]
            new_expr = Operator(expr.name, newargs)
            return (new_expr, [flatarg[1] for flatarg in flat_args])

    elif isinstance(expr, Element):
        if len(expr.args) != 2:
            raise Exception("Only Element expr of type Arr[Var] allowed in objective".format(expr.name))
        flat_idx = flatten_numexpr(expr.args[1])
        flat_arr = [flatten_numexpr(e) for e in expr.args[0]]
        if flat_idx[0] is expr.args[1] and \
           all(arri is flati[0] for (arri, flati) in zip(expr.args[0],flat_arr)):
            return (expr, [])
        else:
            # one of the args was changed
            new_arr = [flati[0] for flati in flat_arr]
            new_expr = Element([new_arr, flat_idx[0]])
            new_cons = flat_idx[1] + [flati[1] for flati in flat_arr]
            return(new_expr, new_cons)

    raise Exception("Expression '{}' not allowed in objective".format(expr)) # or bug
        
def flatten_numexpr(expr):
    """
        input: expression of type:
            * is_num()
            * NumVarImpl
            * Operator with is_type_num()
            * Element with len(args)==2
        output: tuple (base_expr, base_cons) with:
            base_expr one of:
                * is_num()
                * NumVarImpl
                * Operator with is_type_num(), all args are: is_num() or NumVarImpl
                * Operator 'sum', all args are: is_num() or NumVarImpl or Operator '*'[is_num(), NumVarImpl]
                * Element with len(args)==2 and args = ([NumVar], NumVar)
            base_cons: list of flattened constraints (with flatten_constraint(con))
    """
    if is_num(expr) or isinstance(expr, NumVarImpl):
        return (expr, [])

    basecons = []
    if isinstance(expr, Operator):
        # only Numeric operators allowed
        # unary int: '-', 'abs'
        # binary int: 'sub', 'mul', 'div', 'mod', 'pow'
        # nary int: 'sum'
        if expr.is_bool():
            raise Exception("Operator '{}' not allowed as numexpr".format(expr.name))

        # TODO: actually need to flatten THIS expression (and recursively the arguments)
        return (expr, []) # TODO
        newargs = [flatten_numexpr(e) for e in expr.args]


    elif isinstance(expr, Comparison):
        #allowed = {'==', '!=', '<=', '<', '>=', '>'}
        return (expr, []) # TODO

    elif isinstance(expr, Element):
        # A0[A1] == A2
        return (expr, []) # TODO

    # rest: global constraints
    else:
        return (expr, []) # TODO


def is_flatten_var(arg):
    #TODO: extend definition to consider Arithmetic operator expression as valid,
    # e.g. (2a+c-b)/d 
    return is_num(arg) or isinstance(arg, NumVarImpl)

def is_nested_expr(subexpr):
    if is_flatten_var(subexpr):
        return False
    return not all([is_flatten_var(arg) for arg in subexpr.args])

def check_lincons(subexpr, opname):
    LHS, RHS = subexpr.args
    # LHS or RHS is a Var (IntVar, BoolVar, Num)
    if not (is_flatten_var(LHS) or is_flatten_var(RHS)):
        return False 
    
    # LHS is a Var :: RHS does not have nested expr
    elif is_flatten_var(LHS):
        if is_flatten_var(RHS):
            return True 
        # RHS is an LinExp and that's ok
        # as long as it does not have more nested expression 
        if any([is_nested_expr(arg) for arg in RHS.args]):
            return False
        return True

    # Same logic if RHS is a Var
    #TODO: Refactor to avoid code dupcliation
    elif is_flatten_var(RHS):
        if is_flatten_var(LHS):
            return True 
        if any([is_nested_expr(arg) for arg in LHS.args]):
            return False
        return True

    return False 


def flatten_boolexpr(subexpr):
    """
        input: expression of type:
            * True/False
            * BoolVar
            * Operator with is_type_bool()
            * Comparison
        output: tuple (base_expr, base_cons) with:
            base_expr one of:
                * True/False
                * BoolVar
                * Operator with is_type_bool() EXCEPT '->', all args are: BoolVar
            base_cons: list of flattened constraints (with flatten_constraint(con))
    """

    if isinstance(subexpr, BoolVarImpl) or isinstance(subexpr, bool):
        # base case: single boolVar
        return (subexpr, [])

    args = subexpr.args
    base_cons = []

    if isinstance(subexpr, Comparison):
        allowed = {'>','<','<=','>=','==','!='}
        if any([check_lincons(subexpr, opname) for opname in allowed]):
            # Base case: already in base constraint form
            if is_num(args[0]) or is_num(args[1]):
                bvar = BoolVarImpl()
                base_cons += [subexpr == bvar]
            else:
                bvar = args[0] if is_flatten_var(args[0]) else args[1]
                base_cons += [subexpr]
            return (bvar, base_cons)

        # recursive calls to LHS, RHS
        (var1, bco1) = flatten_boolexpr(args[0])
        (var2, bco2) = flatten_boolexpr(args[1])
        bvar = BoolVarImpl()
        base_cons += [bco1, bco2]
        base_cons += [Comparison(subexpr.name, var1, var2) == bvar]
        return (bvar, base_cons)

    elif isinstance(subexpr, Operator):
        # apply De Morgan's transform for "implies"
        if subexpr.name is '->':
            return flatten_boolexpr(Operator('or', -args[0], args[1]))

        if isinstance(subexpr.args[0], BoolVarImpl) and isinstance(subexpr.args[1], BoolVarImpl):
            bvar = BoolVarImpl()
            base_cons += [subexpr == bvar]
            return (bvar, base_cons)

        # nested AND, OR
        # recurisve function call to LHS and RHS
        # XXX: merge nested AND at top level
        (var1, bco1) = flatten_boolexpr(args[0])
        (var2, bco2) = flatten_boolexpr(args[1])
        bvar = BoolVarImpl()
        base_cons += [bco1, bco2]
        base_cons += [Operator(subexpr.name, [var1, var2]) == bvar]
        return (bvar, base_cons)

    

    



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## ortools_python.py
##
"""
    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        ORToolsPython

    ==================
    Module description
    ==================

    ==============
    Module details
    ==============
"""
from .solver_interface import SolverInterface, SolverStatus, ExitStatus
from ..expressions import Comparison, Expression, Operator, Element
from ..globalconstraints import *
from ..model_tools.get_variables import get_variables
from ..model_tools.flatten_model import *

from itertools import cycle

def zipcycle(vars1, vars2):
    v1 = [vars1] if not is_any_list(vars1) else vars1
    v2 = [vars2] if not is_any_list(vars2) else vars2
    return zip(v1, cycle(v2)) if len(v2) < len(v1) else zip(cycle(v1), v2)

class ORToolsPython(SolverInterface):
    """
    Interface to the python 'ortools' API

    Requires that the 'ortools' python package is installed:
    $ pip install ortools

    Creates the following attributes:
    _model: the ortools.sat.python.cp_model.CpModel() created by _model()
    _solver: the ortools cp_model.CpSolver() instance used in solve()
    """

    def __init__(self):
        self.name = "ortools"

    def supported(self):
        try:
            import ortools
            return True
        except ImportError as e:
            return False

    def make_model(self, cpm_model):
        from ortools.sat.python import cp_model as ort

        # Constraint programming engine
        self._model = ort.CpModel()

        # transform into flattened model
        flat_model = flatten_model(cpm_model)

        # Create corresponding solver variables
        self.varmap = dict() # cppy var -> solver var
        modelvars = get_variables(flat_model)
        for var in modelvars:
            if isinstance(var, BoolVarImpl):
                revar = self._model.NewBoolVar(str(var.name))
            elif isinstance(var, IntVarImpl):
                revar = self._model.NewIntVar(var.lb, var.ub, str(var.name))
            self.varmap[var] = revar

        # post the (flat) constraint expressions to the solver
        for con in flat_model.constraints:
            self.post_expression(con)

        # the objective
        if flat_model.objective is None:
            pass # no objective, satisfaction problem
        else:
            obj = self.convert_expression(flat_model.objective)
            if flat_model.objective_max:
                self._model.Maximize(obj)
            else:
                self._model.Minimize(obj)

        return self._model

    def solve(self, cpm_model, num_workers=1):
        if not self.supported():
            raise "Install the python 'ortools' package to use this '{}' solver interface".format(self.name)
        from ortools.sat.python import cp_model as ort

        # store original vars (before flattening)
        original_vars = get_variables(cpm_model)

        # create model
        self.ort_model = self.make_model(cpm_model)
        # solve the instance
        self.ort_solver = ort.CpSolver()
        self.ort_solver.parameters.num_search_workers = num_workers # increase for more efficiency (parallel)
        self.ort_status = self.ort_solver.Solve(self.ort_model)

        # translate status
        my_status = SolverStatus()
        my_status.solver_name = self.name
        if self.ort_status == ort.FEASIBLE:
            my_status.exitstatus = ExitStatus.FEASIBLE
        elif self.ort_status == ort.OPTIMAL:
            my_status.exitstatus = ExitStatus.OPTIMAL
        elif self.ort_status == ort.INFEASIBLE:
            my_status.exitstatus = ExitStatus.UNSATISFIABLE
        else:
            raise NotImplementedError # a new status type was introduced, please report on github
        my_status.runtime = self.ort_solver.WallTime()

        if self.ort_status == ort.FEASIBLE or self.ort_status == ort.OPTIMAL:
            # fill in variables
            for var in original_vars:
                var._value = self.ort_solver.Value(self.varmap[var])

        return my_status

    # for subexpressions (variables, lists and linear expressions)
    def convert_expression(self, expr):
        # python constants
        if is_num(expr):
            return expr

        # list
        if is_any_list(expr):
            return [self.convert_expression(e) for e in expr]

        # decision variables, check in varmap
        if isinstance(expr, NumVarImpl): # BoolVarImpl is subclass of NumVarImpl
            return self.varmap[expr]

        # ~B
        # XXX should we implement NegBoolView instead?
        if isinstance(expr, Comparison) and expr.name == '==' and \
           isinstance(expr.args[0], BoolVarImpl) and expr.args[1] == 0:
            return self.varmap[expr.args[0]].Not()

        if isinstance(expr, Operator):
            # bool: 'and'/n, 'or'/n, 'xor'/n, '->'/2
            # unary int: '-', 'abs'
            # binary int: 'sub', 'mul', 'div', 'mod', 'pow'
            # nary int: 'sum'
            args = [self.convert_expression(e) for e in expr.args]
            if expr.name == 'and':
                return all(args)
            elif expr.name == 'or':
                return any(args)
            elif expr.name == 'xor':
                raise Exception("or-tools translation: XOR probably illegal as subexpression")
            elif expr.name == '->':
                # when part of subexpression: can not use .OnlyEnforceIf() (I think)
                # so convert to -a | b
                return args[0].Not() | args[1]
            elif expr.name == '-':
                return -args[0]
            elif expr.name == 'abs':
                return abs(args[0])
            if expr.name == 'sub':
                return args[0] - args[1]
            elif expr.name == 'mul':
                return args[0] * args[1]
            elif expr.name == 'div':
                return args[0] / args[1]
            elif expr.name == 'mod':
                return args[0] % args[1]
            elif expr.name == 'pow':
                return args[0] ** args[1]
            elif expr.name == 'sum':
                return sum(args)

        raise NotImplementedError # should not reach this... please report on github
        # there might be an Element expression here... need to add flatten rule then?

    def post_expression(self, expr):
        # base case
        if isinstance(expr, BoolVarImpl):
            print(type(expr), expr, self.varmap[expr])
            self._model.AddBoolOr( [self.varmap[expr]] )
        
        # standard expressions: comparison, operator, element
        if isinstance(expr, Comparison):
            # recursively convert arguments (subexpressions)
            args = [self.convert_expression(e) for e in expr.args]
            #allowed = {'==', '!=', '<=', '<', '>=', '>'}
            #XXX refactor decomposition into constructor of Comparison()?
            for lvar, rvar in zipcycle(args[0], args[1]):
                if expr.name == '==':
                    # XXX this might break for 'reification' of constraints...
                    # will have to add two-sided .OnlyEnforceIf() then?
                    # if you get an error here, please report on github
                    self._model.Add(lvar == rvar)
                elif expr.name == '!=':
                    self._model.Add( lvar != rvar )
                elif expr.name == '<=':
                    self._model.Add( lvar <= rvar )
                elif expr.name == '<':
                    self._model.Add( lvar < rvar )
                elif expr.name == '>=':
                    self._model.Add( lvar >= rvar )
                elif expr.name == '>':
                    self._model.Add( lvar > rvar )

        elif isinstance(expr, Operator):
            # bool: 'and'/n, 'or'/n, 'xor'/n, '->'/2
            # unary int: '-', 'abs'
            # binary int: 'sub', 'mul', 'div', 'mod', 'pow'
            # nary int: 'sum'

            # two special cases:
            #    '->' with .onlyEnforceIf()
            #    'xor' does not have subexpression form
            # all others: add( subexpression )
            if expr.name == '->':
                # XXX we might have to do some sort of flattening here if RHS is not BoolVar...
                # or-tools will raise error if invalid LHS (or RHS)
                args = [self.convert_expression(e) for e in expr.args]
                self._model.Add( args[0] ).OnlyEnforceIf(args[1])
            elif expr.name == 'xor':
                args = [self.convert_expression(e) for e in expr.args]
                self._model.AddBoolXor(args)
            else:
                self._model.Add( self.convert_expression(expr) )

        elif isinstance(expr, Element):
            # A0[A1] == A2 --> AddElement(A1, A0, A2)
            args = [self.convert_expression(e) for e in expr.args]
            return self._model.AddElement(args[1], args[0], args[2])
        

        # rest: global constraints
        elif expr.name == 'alldifferent':
            args = [self.convert_expression(e) for e in expr.args]
            self._model.AddAllDifferent(args) 
        elif expr.name == 'min' or expr.name == 'max':
            args = [self.convert_expression(e) for e in expr.args]
            lb = min(a.lb() if isinstance(arg, NumVarImpl) else a for a in args)
            ub = max(a.ub() if isinstance(arg, NumVarImpl) else a for a in args)
            aux = self._model.NewIntVar(lb, ub, "aux")
            if expr.name == 'min':
                self._model.AddMinEquality(aux, args) 
            else:
                self._model.AddMaxEquality(aux, args) 

        else:
            # global constraint not known, try generic decomposition
            dec = expr.decompose()
            if not dec is None:
                self.post_expression(flatten_constraint(dec))
            else:
                raise NotImplementedError # if you reach this... please report on github
        




from .variables import *
from .expressions import *
from itertools import chain, combinations

"""
    ===============
    List of classes
    ===============
    .. autosummary::
        :nosignatures:

        alldifferent
        allequal
        circuit
        GlobalConstraint

    ================
    List of functions
    =================
    .. autosummary::
        :nosignatures:

        min
        max

    ==================
    Module description
    ==================

    Global constraint definitions

    A global constraint is nothing special in CPMpy. It is just an
    expression of type `GlobalConstraint` with a name and arguments.

    You can define a new global constraint as simply as:

    ```
    def my_global(args):
        return GlobalConstraint("my_global", args)
    ```


    Of course, solvers may not support a global constraint
    (if it does, it should be mapped to the API call in its SolverInterface)

    You can provide a decomposition for your global constraint through
    the decompose() function.
    To overwrite it, you should define your global constraint as a
    subclass of GlobalConstraint, rather then as a function above.
    
    Your decomposition function can use any standard CPMpy expression.
    For example:

    ```
    class my_global(GlobalConstraint):
        def __init__(self, args):
            super().__init__("my_global", args)
    
        def decompose(self):
            return [self.args[0] != self.args[1]] # your decomposition
    ```


    If you are modeling a problem and you want to use another decomposition,
    simply overwrite the 'decompose' function of the class, e.g.:

    ```
    def my_circuit_decomp(self):
        return [self.args[0] == 1] # does not actually enforce circuit
    circuit.decompose = my_circuit_decomp # attach it, no brackets!

    vars = IntVars(1,9, shape=(10,))
    constr = circuit(vars)

    Model(constr).solve()
    ```

    The above will use 'my_circuit_decomp', if the solver does not
    natively support 'circuit'.
"""


# min: listwise 'min'
def min(iterable):
    """
        min() overwrites python built-in,
        checks if all constants and computes np.min() in that case
    """
    if not any(isinstance(elem, Expression) for elem in iterable):
        return np.min(iterable)
    return GlobalConstraint("min", list(iterable))

def max(iterable):
    """
        max() overwrites python built-in,
        checks if all constants and computes np.map() in that case
    """
    if not any(isinstance(elem, Expression) for elem in iterable):
        return np.max(iterable)
    return GlobalConstraint("max", list(iterable))



class alldifferent(GlobalConstraint):
    """ all arguments have a different (distinct) value
    """
    def __init__(self, args):
        super().__init__("alldifferent", args)
    
    def decompose(self):
        return [var1 != var2 for var1, var2 in _all_pairs(self.args)]


class allequal(GlobalConstraint):
    """ all arguments have the same value
    """
    def __init__(self, args):
        super().__init__("allequal", args)
    
    def decompose(self):
        return [var1 == var2 for var1, var2 in _all_pairs(self.args)]


class circuit(GlobalConstraint):
    def __init__(self, args):
        super().__init__("circuit", args)
    
    def decompose(self):
        """
            TODO needs explanation/reference
        """
        n = len(self.args)
        a = self.args
        z = IntVar(0, n-1, n)
        constraints = [alldifferent(z),
                       alldifferent(a),
                       z[0]==a[0],
                       z[n-1]==a[0]]
        for i in range(1,n-1):
            constraints += [z[i] != 0,
                            z[i] == a[z[i-1]]]
        return constraints


def _all_pairs(args):
    """ internal helper function
    """
    pairs = list(combinations(args, 2))
    return pairs




"""
    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        SolverInterface
        SolverStatus
        ExitStatus

    ==================
    Module description
    ==================
    Contains the abstract class `SolverInterface` for defining solver interfaces,
    as well as a class `SolverStatus` that collects solver statistics,
    and the `ExitStatus` class that represents possible exist statuses.

    Each solver has its own class that inherits from `SolverInterface`.

"""
#
#==============================================================================
from enum import Enum
import time

#==============================================================================
class SolverInterface(object):
    """
        Abstract class for defining solver interfaces. All classes implementing
        the ``SolverInterface``
    """
    def __init__(self):
        self.name = "dummy"

    def supported(self):
        """
            Check for support in current system setup. Return True if the system
            has package installed or supports solver, else returns False.

        Returns:
            [bool]: Solver support by current system setup.
        """
        return False

    def solve(self, model):
        """
            Build the CPMpy model into solver-supported model ready for solving
            and returns the solver statistics generated during model solving.

        :param model: CPMpy model to be parsed.
        :type model: Model

        :return: an object of :class:`SolverStatus`
        """
        return SolverStatus()

#
#==============================================================================
class ExitStatus(Enum):
    """
        Exit status of the solver

        Attributes:
            NOT_RUN: Has not been run
            OPTIMAL: Optimal solution to an optimisation problem found
            FEASIBLE: Feasible solution to a satisfaction problem found,
                      or feasible (but not proven optimal) solution to an
                      optimisation problem found
            UNSATISFIABLE: No satisfying solution exists
            ERROR: Some error occured (solver should have thrown Exception)
    """
    NOT_RUN = 1
    OPTIMAL = 2
    FEASIBLE = 3
    UNSATISFIABLE = 4
    ERROR = 5


#
#==============================================================================
class SolverStatus(object):
    """
        Status and statistics of a solver run
    """
    exitstatus: ExitStatus
    runtime: time

    def __init__(self):
        self.exitstatus = ExitStatus.NOT_RUN
        self.runtime = None
        self.solver_name = None

    def __repr__(self):
        return "{} ({} seconds)".format(self.exitstatus, self.runtime)



from ..model import *
from ..expressions import *
from ..variables import *

"""
 Model transformation, read-only
 Returns an (ordered by appearance) list of all variables in the model
"""
def get_variables(model):
    # want an ordered set. Emulate with full list that is uniquified
    vars_ = vars_expr(model.constraints)

    # mimics an ordered set, manually...
    return uniquify(vars_)

# https://stackoverflow.com/questions/480214/how-do-you-remove-duplicates-from-a-list-whilst-preserving-order
def uniquify(seq):
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]

def vars_expr(expr):
    # a var, do our thing
    if isinstance(expr, NumVarImpl):
        return [expr]

    vars_ = []
    # if list or Expr: recurse
    if is_any_list(expr):
        for subexpr in expr:
            vars_ += vars_expr(subexpr)
    elif isinstance(expr, Expression):
        for subexpr in expr.args:
            vars_ += vars_expr(subexpr)
    # else: every non-list, non-expression
    return vars_



from .minizinc_python import MiniZincPython
from .minizinc_text import MiniZincText
from .ortools_python import ORToolsPython

def get_supported_solvers():
    """
        Returns a list of currently supported solvers.

    :return: a list of SolverInterface sub-classes :list[SolverInterface]:
    """
    return [sv for sv in builtin_solvers if sv.supported()]

# Order matters! first is default, then tries second, etc...
builtin_solvers=[ORToolsPython(),MiniZincPython()]



# Tias Guns, 2019
# current version
#==============================================================================
# VERSION = (0, 5, 2, "dev")

# __version__ = "%d.%d.%d.%s" % VERSION if len(VERSION) == 4 else \
#               "%d.%d.%d" % VERSION
__version__ = "0.5.3"

from .variables import *
from .expressions import *
from .globalconstraints import *
from .model import *
